function u = propagate2(u0, d, h, dt, numt, filename)
% Local level set method propagation
%   u0 - initial implicit function
%   d  - distance function
%   h  - spatial subdivision: dx = dy = dz = h
% Note: we will use both h = 1/(N-1) and N to favour multiplication over addition

initderivatives;

gd = grad(d);

u = u0;

% for convenience construct a part of the propagation as a mtx
s = zeros(size(u));

gamma = 4*h;
beta = 2*h;

[mask idx] = gennarrowband(u, gamma);
[gu ngu meangu] = gradnarrow(u, gd, mask, idx);

% allocate space for mask array of the narrow band
for n = 1:numt
  E = energynarrow(u, d, nD, h, mask, idx);
  Einv = 1/E;

  maxspeed = 0.0;
  % propagate the narrow band interface by one time step
  [xn yn zn] = size(u);
  u1 = zeros(xn, yn, zn); % used in 2 stage Runge Kutta scheme
  L = zeros(xn, yn, zn);
  for l = idx
    i = l(1); j = l(2); k = l(3);
    if mask(i,j,k) == 2
      % Compute the speed of propagation
      au = abs(u(i,j,k));
      if au < beta
        cutoff = 1;
      elseif au < gamma
        cutoff = (au - gamma)*(au - gamma)*(2*au - 2*h)/(8*h*h*h);
      else
        cutoff = 0;
      end
      Kngu = curvature(u, i, j, k, D);
      if isnan(Kngu) || Kngu == inf
        fprintf('1. curvature is nan at %d, %d, %d\n', i, j, k);
        return
      end
      prop = gd(:,i,j,k)'*gu(:,i,j,k) + 0.5*d(i,j,k)*Kngu;
      if isnan(prop) || prop == inf
        fprintf('1. prop is nan at %d, %d, %d\n', i, j, k);
        return
      end
      L(i,j,k) = cutoff*Einv*d(i,j,k)*prop;
      if isnan(L(i,j,k)) || L(i,j,k) == inf
        fprintf('1. L is nan at %d, %d, %d\n', i, j, k);
        return
      end
      speed = L(i,j,k)/ngu(i,j,k);
      if speed > maxspeed 
        maxspeed = speed;
      end
    end % if mask == 2
  end % for each i

  dt = 0.5*h/maxspeed; % adaptive time step heuristic

  for l = idx
    i = l(1); j = l(2); k = l(3);
    if mask(i,j,k) == 2
      u1(i,j,k) = u(i,j,k) + dt*L(i,j,k); % update to first stage of RK
    end
  end

  [mask1 idx1] = gennarrowband(u1, gamma);
  [gu1 ngu1] = gradnarrow(u1, gd, mask1, idx1);
  E1 = energynarrow(u1, d, nD, h, mask1, idx1);
  E1inv = 1/E1;

  maxspeed = 0.0;
  % second stage of Runge-Kutta
  for l = idx1
    i = l(1); j = l(2); k = l(3);
    if mask1(i,j,k) == 2
      % Compute the speed of propagation
      au = abs(u1(i,j,k));
      if au < beta
        cutoff = 1;
      elseif au < gamma
        cutoff = (au - gamma)*(au - gamma)*(2*au - 2*h)/(8*h*h*h);
      else
        cutoff = 0;
      end
      Kngu = curvature(u1, i, j, k, D);
      if isnan(Kngu) || Kngu == inf
        fprintf('2. curvature is nan at %d, %d, %d\n', i, j, k);
        u = u1;
        return
      end
      prop = gd(:,i,j,k)'*gu1(:,i,j,k) + 0.5*d(i,j,k)*Kngu;
      if isnan(prop) || prop == inf
        fprintf('2. prop is nan at %d, %d, %d\n', i, j, k);
        u = u1;
        return
      end
      L2 = cutoff*E1inv*d(i,j,k)*prop;
      if isnan(L2) || L2 == inf
        fprintf('2. L is nan at %d, %d, %d\n', i, j, k);
        u = u1;
        return
      end
      speed = 0.5*(L(i,j,k) + L2)/ngu1(i,j,k);
      if speed > maxspeed
        maxspeed = speed;
      end
      u(i,j,k) = u(i,j,k) + 0.5*dt*(L(i,j,k) + L2); % update u
    end % if mask == 2
  end % for each i

  if dt > h/maxspeed 
    fprintf('WARNING: adaptive time step heuristic failed, CFL condition unsatisified: dt = %f > %f !\n', dt, h/maxspeed);
  end

  [mask idx] = gennarrowband(u, gamma);
  [gu ngu meangu] = gradnarrow(u, gd, mask, idx);
  fprintf('mean |Du| = %f\n', meangu);

  % Reinitialize u
  steps = 1;
  while meangu > 1.3 && steps < 10 % max 10 reinitialization steps
    u = reinitialize(u, nD, h, mask, idx);
    [gu ngu meangu] = gradnarrow(u, gd, mask, idx);
    steps = steps + 1;
  end % while meangu is big
  if steps > 1
    fprintf('mean |Du| = %f after reinitialization\n', meangu);
  end

  if nargin > 5 && mod(n,5) == 0
    clf
    dispoutput(u, h, [filename num2str(n)]);
  end
end % for each n

